
/*************************************************************************/
/*
		BANCOMM bc63{57}VME/ bc35{07}VXI Time and Frequency Processor Driver

		for HP-RT v1.1X Systems

		Copyright (C) 1994, Datum Inc, Bancomm Div. All Rights Reserved.

		Author: Greg Dowd
		Datum Inc., Bancomm Div.
		6541 Via Del Oro
		San Jose, CA 95119-1294
		(408)578-4161
*/
/************************************************************************/
/*
 * v1.0a 6/15/95 gdowd
 * fixed error in send_packet which was terminating if 0 in data
 *
 * v1.0b 9/19/00 seffertz
 * fixed error in interrupt handling for multiple boards
*/
#include "./btfpdrvr.h"

#define MAXBUF 1024
static char hold[MAXBUF];
static void btfp_int();
int send_packet();

/****************************************************************************/
/*	btfp device state information structures.  MAXDEV is defined in
	btfpd.h as the max number of bc637's this driver can support.
*/

static struct btfp_status {	
		int intwait;				/* Copy of MASK interrupt enable bits.*/
		int	signalon;				/* True if signal sending is enabled.*/
		int	packeton;				/* True if packet mode is enabled.*/
		int	debugon;				/* True if debug mode is enabled.*/
		int procp;					/* Pointer to calling process */
		int gps_synch;				/* True if synched with GPS packets */
		int rpacki;					/* Read Packet buffer index.			*/
		int wpacki;					/* Write Packet buffer index.			*/
		unsigned char rpack_buffer[MAXPACKET+1];	/* used for read calls */
		unsigned char wpack_buffer[MAXPACKET+1];	/* used for write calls */
		unsigned char work_buffer[MAXPACKET+1];		/* used to build packet */
} btfp_status[ MAXDEV ];				/* One for each bc63{57} supported.	 */

/************************************************************************/
/*
	btfpinstall - called during power-up

					Try to access each btfp defined, up to
					MAXDEV, which is our local definition
					of how many btfp's this driver can support.
					If an access succeeds, mark the board		
					on-line and announce it is there, else mark it		
					off-line and announce it is not there.  'Opens' to 
					boards either not there or beyond MAXDEV will fail.
					
					For boards that are there, their interrupt registers
					are loaded so that a subsequent interrupt enable (via
					a WMASK ioctl(2) call which turns on any bits) will
					allow an interrupt to come in.  The vector and
					the interrupt level are obtained from the btfpinfo.h
					initialized during kernel configuration.
																		
*/                                                                      
/************************************************************************/

/* Allocate memory for static information saved by each device */
/*  Return an error if the call is not successful */  
/* Start the system thread to handle theinterrupts */
/* The interrupt handler should be set up here. */
/* Turn on interrupts for this particular interrupt bit */
bcd *btfpinstall(info)
struct bc_info *info;
{
	bcd *DrvrData;				/* struct passed to driver entry points */	
	int ps;						/* rtn value from iointset call			*/
	unsigned short *BCva;		/* pointer to board virtual address		*/
	register BTFPVOLATILE *dp;				/* Pointer to board base reg.	*/
	register struct card_info *cp;			/* Pointer to card_data init	*/
	register num_dev;						/* # devices to initialize.		*/
	register i;								/* Index to a btfp & strucs.	*/
	register j;								/* Loop index for inits.		*/
	ushort	temp;							/* Temp place to access board.	*/

	go_away_ifnot_this_cpu(BTFP0_VME_VEC);
	if ((DrvrData = (bcd *)sysbrkequiv ((long)(sizeof (bcd)),
	     DEF_SBRK_ALIGN)) == NULL )
	{						/* allocate shared mem for drvr data struct */
		pseterr(ENXIO) ;
		return((bcd *)SYSERR) ;
	}

	if( BTFPCOUNT > MAXDEV )				/* Support only MAXDEV devices.	*/
	{
		num_dev = MAXDEV;					/* Truncate count to the max.	*/
		kprintf("btfpinit(): Driver supports only the first ",
					"%d btfp configured.\n", MAXDEV );
	}
	else									/* Else we can support all of	*/
	{
		num_dev = BTFPCOUNT;				/* the btfp's configured.		*/
	}

	for (i = 0; i < num_dev; i++) 			/* their structures.			*/
	{	
		dp = (BTFPVOLATILE *) sysshmem_to_va(A16+DATA_ACCESS,
			(char *)info[i].btfp_physaddr);	/* Pointer to base register.	*/

		cp = &DrvrData->card_data[i];		/* Pointer to card_info struct	*/

		dp->btfp_mask   = 0x00;
		dp->btfp_vector = info[i].btfp_vector;
		dp->btfp_level  = info[i].btfp_level;
		btfp_status[ i ].intwait  = 0;	/* NOT waiting for an int.*/
		btfp_status[ i ].signalon = 0;	/* Signal mechanism is turned OFF.*/
		btfp_status[ i ].packeton = 0;	/* Packet mechanism is turned OFF.*/
		btfp_status[ i ].debugon  = 0;	/* Debug mode is turned OFF.	*/
		btfp_status[ i ].gps_synch= 0;	/* NOT synched with GPS packets. */
		cp->vector = &info[i].vector;	/* define driver interrupt vector	*/
		cp->int_info.interrupt_action = SIGNAL_ONLY ;
										/* define driver interrupt action */
										/* as SIGNAL_ONLY.  This lets the */
										/* int handler run as a normal    */
										/* kernel thread.				  */
		cp->int_info.sem = &DrvrData->card_data[i].bcdrvr_sem ;
										/* set pointer to semaphore which */
										/* kernel will ssignal() on int   */
		cp->btfp_level  = info[i].btfp_level;
										/* define btfp interrupt level	*/
		DrvrData->card_data[i].btfp_vector = info[i].btfp_vector;
										/* define btfp interrupt vector	*/
		DrvrData->card_data[i].bcdrvr_sem = 0 ;	/* initialize semaphore to 0  */
		DrvrData->card_data[i].pBC = dp;  		/* set ptr to board virt addr */
		DrvrData->card_data[i].user_sem = 0 ;	/* initialize user semaphore  */
		
		/* Create the interrupt handler thread */
		if(ststart(btfp_int,MINSTK,51,"btfp_int",2,DrvrData,i) == SYSERR)
		{
			pseterr(EIO);
			kprintf("\nCouldn't start int thread\n") ;
			return((bcd *)SYSERR) ;
		}

		/* Set the interrupt  handler */
		if((ps = iointset(DrvrData->card_data[i].vector,
						(char *)NULL ,&DrvrData->card_data[i])) == SYSERR)
		{
			kprintf("\nCouldn't setup the interrupt\n") ;
			return((bcd *)SYSERR) ;
		}

		/* Enable our interrupt */
		dvrenable(DrvrData->card_data[i].vector->interrupt_bit) ; 

	}


	return(DrvrData) ;
}

btfpopen(DrvrData, mmdev, fd)
bcd *DrvrData;
int mmdev;
struct file *fd;
{
	int ps,ps2;
	register BTFPVOLATILE *dp = (BTFPVOLATILE *)
								DrvrData->card_data[ minor(mmdev) ].pBC;
										/* Init pointer to base register */
	register struct btfp_status *sp = &btfp_status[minor(mmdev)];
										/* Init pointer to status register. */
	ushort	ustemp, usmask;				/* For use turning off interrupts. */
	int		s;							/* Interrupt level from spl().  	*/

	if(( minor(mmdev) > MAXDEV) || (dp == (BTFPVOLATILE *)0))
	{
		pseterr(ENXIO);
		return SYSERR;					/* Set no such device or addr error */
	}
	/* Do a successful, normal OPEN */

	dvrdisable(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
											/* disable board interrupt	*/
	sp->intwait = 0;						/* Indicate not waiting for int. */
	sp->signalon = 0;						/* Signal mechanism turned off   */
	sp->packeton = 0;						/* Packet mode is turned off.    */
	sp->debugon  = 0;						/* Debug mode is turned off.     */

	dp->btfp_mask = 0;						/* Turn off interrupt.           */

	usmask = LOCKDIS | HBDIS | EVENTDIS | STRDIS; 
											/* Disable lockout, heartbeat    */
											/* event & strobe interrupts.    */
	dp->btfp_cr0 = usmask;
	sp->procp = currpptr;

	dvrrestore(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
											/* reenable board interrupt.     */
	return(0);								/* Return open successful.       */
}
  
btfpclose(DrvrData, fd)
bcd *DrvrData;
struct file *fd;
{
	int mmdev = fd->dev;

	register BTFPVOLATILE *dp = (BTFPVOLATILE *)
								DrvrData->card_data[ minor(mmdev) ].pBC;
										/* Init pointer to base register */
	register struct btfp_status *sp = &btfp_status[minor(mmdev)];
										/* Init pointer to status register. */
	ushort	ustemp, usmask;				/* For use turning off interrupts. 	*/
	int		ps;							/* Interrupt restore value.		  	*/

	if(( minor(mmdev) > MAXDEV) || (dp == (BTFPVOLATILE *)0))
	{
		pseterr(ENXIO);
		return SYSERR;					/* Set no such device or addr error */
	}
	/* Do a successful, normal CLOSE */

	dvrdisable(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
											/* disable board interrupt	*/
	sp->intwait  = 0;						/* Indicate not waiting for int. */
	sp->signalon = 0;						/* Signal mechanism turned off   */
	sp->packeton = 0;						/* Packet mode is turned off.    */
	sp->debugon  = 0;						/* Debug mode is turned off.     */

	dp->btfp_mask = 0;						/* Turn off interrupt.           */

	usmask = LOCKDIS | HBDIS | EVENTDIS | STRDIS; 
											/* Disable lockout, heartbeat    */
											/* event & strobe interrupts.    */
	dp->btfp_cr0 = usmask;
	
	sp->procp = 0;							/* Pointer to user process for	*/
											/* sending signals				*/

	dvrrestore(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);	
											/* reenable board interrupt.    */
	return(0);								/* Return close successful.     */
}

/************************************************************************/
/*
	btfpread()- Called when an application issues a read(2) call.


	(default) PACKETOFF mode:

	Set to Free Running Mode (mode 1), then makes a time request.

	Access the bc63{57}'s capture register to freeze the time,
	then return (converting BCD to ASCII) bytes of the time
	to the caller, until the number of bytes he requested is
	reached or until we run out of bytes to return.

	Sends 34 bytes maximum, separating each of the 16 bytes
	of time data with 3 spaces, and new-line and null
	terminating the string.

	This routine exits when 34 bytes have been sent in response
	to a read, regardless of what u.u_count is.  If it was not
	zero, then presumably the caller (e.g. cat()) will call
	again, but with u.u_offset = 33, which is our indication to
	just return without any transfer, which will result in the
    caller seeing 0 bytes transferred, which means EOF, by
	UNIX definition.


	SIGNALON/SIGNALOFF modes have no effect on this mode.

	PACKETON mode:

	Block until a data packet is available from the OUTFIFO (unless
	SIGNALON mode), then fill the buffer supplied by the read call
	with the data from the OUTFIFO.

		SIGNALON notes:

		If both PACKETON and SIGNALON modes are in effect, then
		the user must issue the ioctl calls to clear the
		DPA interrupt status (WINTSTAT) and enable DPA interrupts
		in the MASK register in a SIGNALON (non-blocking) WMASK ioctl call.
		When the signal is received to indicate the DPA interrupt has
		occurred, the signal handler should issue the read(2) call,
		still leaving the driver in the SIGNALON mode. This will
		cause the btfpread() routine to immediately read the data
		packet from the OUTFIFO.

*/
/************************************************************************/

#define MAXTBYTES	49
#define MAXABYTES	22

btfpread(DrvrData, fd, buff, count)
bcd *DrvrData;
struct file *fd;
char *buff;
register int count;

{
	 int mmdev = fd->dev;					/* get dev nbr from fd			*/
	 register int pass_count = count;		/* set up counter				*/
	 register j = 0;						/* Loop index					*/

	 register BTFPVOLATILE *dp = (BTFPVOLATILE *)
									DrvrData->card_data[ minor(mmdev) ].pBC;

	 register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										 /* Init pointer to status structure.*/
																					 unsigned short dummy;                   /* Dummy for use accessing captr.*/
	 unsigned short ustemp, ustemp2;         /* Temps for use accessing regs.*/
	 char timearray[ MAXABYTES ];            /* Place to put converted time. */
	 register i = 0;                         /* Loop index thanks to K&R.    */
	 register twi;                           /* Time word index.             */
	 register tai;                           /* Time array index.            */
	 unsigned char lochar;                   /* Place to build the ASCII.    */
	 unsigned char hichar;                   /* Ditto.                       */
	 int rtn;                                /* catcher for return code */
	 int ps;								 /* previous state of int	*/
	 int s;                                  /* return from spl()    */
	 int length;                             /* Used for PACKETON read       */
	 unsigned char *user_buffer;             /* pointer into user buffer */

    union char_union					/* For converting short to 'chars'.*/
	{
		unsigned char   two_chars[ 2 ]; 	/* Two characters holder.       */
		unsigned short  time_word;      	/* One of the time words.       */
	} ch_union;

	if( minor(mmdev) >= MAXDEV )			/* Beyond supportable range.    */
		{
		pseterr(ENXIO);			           /* Return error indication.        */
		return(SYSERR);
		}

	if (!sp->packeton)
	{
	/* PACKETOFF mode (default) */
    dummy = dp->btfp_unlock;                /* Release capture lock if set. */
	dummy = dp->btfp_timereq;               /* Freeze the time via access to
											the timereq register.       */

											/* Unpack bytes from btfp.. */

	timearray[ MAXABYTES - 1 ] = '\0';      /* Set up zero termination.     */
	timearray[ MAXABYTES - 2 ] = '\n';      /* And new-line termination.    */


	tai = MAXABYTES - 3;                    /* Fill time array back to front.*/
	for( twi = 4; twi >= 0; twi--)          /* Read time; BCD->ASCII.       */
	{
		ch_union.time_word = dp->btfp_time[ twi ];
											/* Get a time word from module*/

		lochar = ch_union.two_chars[ 1 ];   /* Get two digits to process.   */
		hichar = ch_union.two_chars[ 1 ];   /* And into another var.        */

		lochar = ( lochar >> 4 ) | 0x30;    /* Convert a nibble to ASCII byte.*/
		hichar = ( hichar & 0xf) | 0x30;    /* Convert the other nibble.    */

		timearray[ tai ] = hichar;          /* Move into the time array.    */
		tai--;                              /* Decrement the time array indx.*/
		timearray[ tai ] = lochar;			/* Move the other byte.			*/
		tai--;								/* Decrement the time array indx.*/

		lochar = ch_union.two_chars[ 0 ];   /* Get two digits to process.   */
		hichar = ch_union.two_chars[ 0 ];   /* And into another var.        */

		lochar = ( lochar >> 4 ) | 0x30;    /* Convert a nibble to ASCII byte.*/
		hichar = ( hichar & 0xf) | 0x30;    /* Convert the other nibble.    */

		timearray[ tai ] = hichar;          /* Move into the time array.    */
		tai--;                              /* Decrement the time array indx.*/
		timearray[ tai ] = lochar;          /* Move the other byte.         */
		tai--;                              /* Decrement the time array indx.*/
		if( tai == 0 ) break;               /* End the loop after MAXTBYTES. */
	}
                                            /* Now send time string to caller.*/
	i = 0;                                  /* Initialize byte index.       */
	while(pass_count)
	{
		if( (i <= MAXABYTES - 1) && (fd->position < MAXTBYTES) ) 
											/* Deliver up to max bytes...	*/
											/* And stop on a later read.	*/
		{
			if( i == 2 ||                   /* Between the time fields...   */
				i == 3 ||                   /* Day, hour, minute..etc...    */
				i == 6 ||
				i == 8 ||
				i == 10 ||
				i == 12 ||
				i == 15 ||
				i == 18 ||
				i == 19)                    /* Insert spaces for readability.*/
			{
			for(j=0;j<3;j++)
			{
				*buff++ = ' ';
				fd->position++;
				pass_count--;
			}
			}
			*buff++ = timearray[i];			/* Move time byte to user buff   */
			fd->position++;					/* increment logical file marker */	
			pass_count--;					/* decrement read request count  */
			i++;                            /* Move on to the next byte.    */
		}
		else                                /* We have no more to give...   */
		{                                   /* Or he doesn't want any more. */
			break;                          /* Exit the while and the read. */
		}
	}
	return(count - pass_count);				/* Return number of bytes passed */
}	/* end of PACKETOFF mode */

else /* start of PACKETON mode */
{
	if(pass_count >= sizeof(sp->rpack_buffer) ) /* Beyond pkt buffer size.	*/
	{
		pseterr(ENXIO);
		return(SYSERR);
	}
	if (! (sp->signalon) )					/* Only block in SIGNALOFF mode	*/
	{

		dvrdisable(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
											/* disable board interrupt	*/
											/* Wait for Data Packet Available */
			if (! (sp->gps_synch) )			/* Are we in synch with GPS?	*/
			{
				dp->btfp_ack = ACK_CLEAR;	/* Clear OUTFIFO			*/
											/* When we get the next		*/
											/* DPA, the interrupt		*/
											/* routine will get us		*/
											/* synch-ed up with GPS		*/
											/* and set the flag			*/
			}
			dp->btfp_ack = ACK_DPA;
			sp->intwait = MASK_DPA;		/* indicate waiting for a DPA int.	*/
			dp->btfp_mask = MASK_DPA;	/* Set MASK reg. for a DPA int.		*/

			dvrrestore(ps,
					DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);

			while ( sp->intwait )		/* btfp_int() will clear this on int. */
			{
										/* block, waiting for 637 interrupt. */
				swait(&DrvrData->card_data[minor(mmdev)].user_sem,0);
										/* Wait till our interrupt occurs. */
			}


	}	/* end of SIGNALOFF section */

	/* Interrupt routine has synched us up with the TANS and read in	*/
	/* the next packet (rpacki bytes) from the OUTFIFO into rpack_buffer */

	if (sp->rpacki <= pass_count)		/* If less bytes than requested	*/
	{
		rtn = memcpy (buff,
						sp->rpack_buffer,
						sp->rpacki);
		if (rtn != buff)
		{
			/* Error */
			kprintf("copyout (partial) error in btfpread\n");
			return(SYSERR);
		}

	return(sp->rpacki);
	} /* end of "emptied our buffer first" case */
	else
	{
		/* user didn't give us a big enough buffer!	*/

		rtn = memcpy(buff,
					  sp->rpack_buffer,
					  pass_count);
		if (rtn != buff)
		{
			/* Error! */
			kprintf("copyout (full) error in btfpread\n");
			return(SYSERR);
		}
	return(pass_count);
	} /* end of "filled user buffer" case */

} /* end of PACKETON mode */

} /* end of btfpread routine */

/************************************************************************/
/*
	btfpwrite() - Called when an application does a write(2) call.

	(default) PACKETOFF mode:

	Sets to Free Running Mode (mode 1), waits for the 1PPS pulse (unless
	in SIGNALON mode), then loads time from user buffer.

	Accepts a buffer in UNIX's "date '%j%H%M%S'" format, namely Julian date,
	hours, minutes, then seconds (with or without intervening spaces). Prepends
	this buffer with a 'B' (to indicate Load Major Time). Note that this routine
	can accept the same format output by the default read(2) call, by
	stripping off the ms, us, and ns fields.

			SIGNALON notes:

			If both PACKETOFF and SIGNALON modes are in effect, then
			the user must issue the ioctl calls to clear the
			1PPS interrupt status (WINTSTAT) and enable 1PPS interrupts
			in the MASK register in a SIGNALON (non-blocking) WMASK ioctl call.
			When the signal is received to indicate the 1PPS interrupt has
			occurred, the signal handler should issue the write(2) call,
            still leaving the driver in the SIGNALON mode. This will
    	    cause the btfpwrite() routine to immediately write the data
		    packet (containing the date and time) to the INFIFO.

   PACKETON mode:

   Accepts a buffer from the user and transmits it directly to the INFIFO.

   SIGNALON/SIGNALOFF modes have no effect on this mode.

/************************************************************************/

btfpwrite(DrvrData, fd, buff, count)
bcd *DrvrData;
struct file *fd;
char *buff;
register int count;
{
	int mmdev = fd->dev;					/* get dev nbr from fd			*/
	register int pass_count = count;		/* set up counter				*/
	        
	 register BTFPVOLATILE *dp = (BTFPVOLATILE *)
									DrvrData->card_data[ minor(mmdev) ].pBC;
											 /* Init pointer to base register.*/
	 /* Pointer to btfp registers.   */
	 register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										 /* Init pointer to status structure.*/
    int rtn;                                /* catcher for return code */
	int s;                                  /* spl() rtn code */
	register int ps;						/* interrupt state */
	unsigned char *user_buffer;             /* pointer into user buffer */
	register worki;                         /* buffer index */
	int num_bytes;                          /* Number of user bytes */
	unsigned short usmask;                  /* Used for register access */

	if( minor(mmdev) >= MAXDEV )                   /* Beyond supportable range.    */
	{
		pseterr(ENXIO);			           /* Return error indication.        */
		return(SYSERR);
	}
	if (!sp->packeton)
	{
	/* PACKETOFF mode(default) 	*/

		num_bytes = count;					/* Save number of user bytes */
											/* Do this now to save time later */
		rtn = memcpy(sp->work_buffer,
					  buff,
					  count);
		if (rtn != sp->work_buffer)
		{
			/* Error !	*/
			kprintf("memcpy (packetoff) Error in btfpwrite\n");
			pseterr(ENXIO);
			return(SYSERR);
		}
		
										/* Leave room for 'B' 	*/
		sp->wpack_buffer[0] = SOH;		/* Make into null-terminated B packet*/
		sp->wpack_buffer[1] = 'B';
		sp->wpacki = 2;
		for (worki = 0; worki <=num_bytes; worki++ )
										/* Eliminate white space, plus */
										/* ms, us, and ns fields		*/
		{
			if ( (sp->work_buffer[worki] != ' ')
				&& (sp->work_buffer[worki] != '\n')
				&& (sp->wpacki <= 10) )		/* (dddhhmmss takes 9 bytes) */
			{
				sp->wpack_buffer[sp->wpacki++] = sp->work_buffer[worki];
			}
		}
		sp->wpack_buffer[sp->wpacki++] = ETB;		/* Finish it off.	*/
		sp->wpack_buffer[sp->wpacki] = (char) NULL;	/* with ETB and NULL*/
		
		if (! sp->signalon)					/* Only block in signaloff mode*/
		{
			dvrdisable(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
													/* Disable interrupt */
			if (sp->debugon)
				kprintf("T");
			sp->intwait = MASK_1PPS;	/* Indicate waiting for 1PPS int. */
			dp->btfp_intstat = INTSTAT_1PPS;	/* Clear 1PPS int. status */
			usmask = dp->btfp_intstat;			/* Clear ALL int. status  */
			dp->btfp_intstat = usmask;			/* ... */
			dp->btfp_mask = MASK_1PPS;			/* Set MASK reg. for 1PPS int.*/
			dvrenable(DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
		
			while ( sp->intwait )		/* btfp_int() will clear this on int. */
			{
										/* block, waiting for 637 interrupt. */
				swait(&DrvrData->card_data[minor(mmdev)].user_sem,0);
										/* Wait till our interrupt occurs. */
			}
	
			/* Interrupt routine should send the packet in wpackbuffer 	*/

		} /* end of (blocking) SIGNALOFF mode */

			dvrdisable(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
												/* block interrupt */
			if ( (send_packet (DrvrData, mmdev, sp->wpack_buffer,sp->wpacki)) == 1)
													/* Load time buffer */
			{
			dvrrestore(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
												/* restore interrupt */
			kprintf("btfpwrite: Error from send_packet of wpack_buffer\n");
			pseterr(EIO);
			return(-3);
			}
			dvrrestore(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
												/* restore interrupt */

			return(count);						/* Return no problem */
		}
		else
		{
		/* PACKETON mode */
			
			if ( count >= sizeof(sp->wpack_buffer) ) /* beyond pkt buffer size*/
			{
				pseterr(ENXIO);
				return(-4);
			}
			num_bytes = count;					/* Save number of user bytes */

			rtn = memcpy(sp->wpack_buffer,
						  buff,
						  count);
			if ( rtn != sp->wpack_buffer )
			{
				/* Error ! */
				kprintf("memcpy (packeton) error in btfpwrite\n");
				pseterr(EIO);
				return(-5);
			}

			sp->wpack_buffer[num_bytes] = (unsigned char) NULL;
											/* Make null terminated string */
			dvrdisable(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
												/* block interrupt */
			/* 6/15/95 gdowd */
			if ( (send_packet (DrvrData, mmdev, sp->wpack_buffer, count)) == 1)
													/* Send raw packet */
			{
			dvrrestore(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
												/* restore interrupt */
			pseterr(EIO);
			return(-6);
			}
			dvrrestore(ps,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
												/* restore interrupt */
			return(count);						/* Return # of bytes written */
		
		}/* end of PACKETON mode */

} /* end of btfpwrite routine */
/************************************************************************/
/*
	  btfpioctl()- Called when an application does an ioctl(2) call.


				  Generally, the bc63{57}'s registers are written or read
				  by these routines.

				  If we are in SIGNALON mode, we do not block when
				  an external interrupt enable (to MASK) is received,
				  else we do block.

				  /// We are enabling the interrupt after the cause
					  of it may have happened, e.g. strobe may already
					  have gone off.  Hence an application would sleep
					  forever.
*/
/************************************************************************/

btfpioctl(DrvrData, fd, cmd,arg)
bcd *DrvrData;
struct file *fd;
int cmd;
char *arg;
{
	int mmdev = fd->dev;					/* define mmdev for btfpioctl */
	register BTFPVOLATILE *dp = (BTFPVOLATILE *)
									DrvrData->card_data[ minor(mmdev) ].pBC;
											/* Init pointer to base register.*/
	/* Pointer to btfp registers.   */
		
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
										/* Init pointer to status structure.*/
	register struct btfp_device *up;
										/* READALL receives such a pointer.*/
	register struct btfp_time *tp;
										/* Pointer to time register structure.  */
	register struct btfp_event *ep;
										/* Pointer to event register structure. */
	register struct btfp_strobe *strp;
										/* Pointer to strobe register structure.    */
	struct btfp_device btfpdev;
	struct btfp_time btfptime;
	struct btfp_event btfpevent;
	struct btfp_strobe btfpstrobe;
	int     s;                      /* Interrupt level from spl().          */
	volatile unsigned short ustemp, ustemp2; /* Temps for use accessing regs         */
	volatile unsigned int uitemp;            /* Temp for use in returning values.    */
	unsigned short  usmask = MASK_INT0 | MASK_INT1 | MASK_INT2 | MASK_INT3 | MASK_INT4;
									/* Mask for turning off device interrupts.*/

	up = &btfpdev;
	tp = &btfptime;
	ep = &btfpevent;
	strp = &btfpstrobe;

	switch (cmd) {                  /* Handle specific incoming commands.   */

    case RIDR:                      /* Read the ID register.        */
		ustemp = dp->btfp_id;
		uitemp = 0x0 | ustemp;      /* Right justify value in an int.       */
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */	
		break;

	case RDTR:                      /* Read the Device Type register.       */
		ustemp = dp->btfp_device;
		uitemp = 0x0 | ustemp;      /* Right justify value in an int.       */
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */	
		break;

	case RSR:                       /* Read the Status register.        */
		ustemp = dp->btfp_status;
		uitemp = 0x0 | ustemp;      /* Right justify value in an int.       */
		memcpy (arg, &uitemp, sizeof(int));    /* Move to caller's space */
		break;

	case WCR:                       /* Write to the Control register.   */
		dp->btfp_control = (ushort)arg;/* Directly from the user's argument. */
		break;

	case RTIMEREQ:					/* Read the Time request register.		*/
   		ustemp = dp->btfp_timereq;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case RTIME0:					/* Read time word zero.					*/
		ustemp = dp->btfp_time[ 0 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case RTIME1:					/* Read time word one.					*/
		ustemp = dp->btfp_time[ 1 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case RTIME2:					/* Read time word two.					*/
		ustemp = dp->btfp_time[ 2 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case RTIME3:					/* Read time word three.				*/
		ustemp = dp->btfp_time[ 3 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case RTIME4:					/* Read time word four.					*/
		ustemp = dp->btfp_time[ 4 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case REVENT0:					/* Read event word zero.				*/
		ustemp = dp->btfp_event[ 0 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case REVENT1:					/* Read event word one.					*/
		ustemp = dp->btfp_event[ 1 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case REVENT2:					/* Read event word two.					*/
		ustemp = dp->btfp_event[ 2 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case REVENT3:					/* Read event word three.				*/
		ustemp = dp->btfp_event[ 3 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case REVENT4:					/* Read event word four.				*/
		ustemp = dp->btfp_event[ 4 ];	
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case WSTROBE1:					/* Write strobe word one.				*/
   		dp->btfp_strobe1 = (ushort)arg;/* Directly from the user's argument.	*/
		break;

	case WSTROBE2:					/* Write strobe word two.				*/
   		dp->btfp_strobe2 = (ushort)arg;/* Directly from the user's argument.	*/
		break;

	case WSTROBE3:					/* Write strobe word three.				*/
   		dp->btfp_strobe3 = (ushort)arg;/* Directly from the user's argument.	*/
		break;

	case RUNLOCK:						/* Read Unlock register.	*/
   		ustemp = dp->btfp_unlock;		/* Contents of the register matter not.	*/
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;
	
	case RACK:							/* Read the ACK register.		*/
   		ustemp = dp->btfp_ack;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case WACK:						/* Write to the ACK register.	*/
   		dp->btfp_ack = (ushort)arg;/* Directly from the user's argument.	*/
		break;

	case WCR0:						/* Write control register zero.		 	*/
		dp->btfp_cr0 = (ushort)arg; 	/* Directly from the user's argument.*/
		break;

	case RCR0:						/* Read control register zero.		 	*/
		ustemp = dp->btfp_cr0;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;
	
	case ROUTFIFO:					/* Read the OUTFIFO register.		*/
   		ustemp = dp->btfp_outfifo;		
		uitemp = 0x0 | ustemp;		/* Right justify value in an int.		*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case WINFIFO:					/* Write to the INFIFO register.	*/
   		dp->btfp_infifo = (ushort)arg;/* Directly from the user's argument.	*/
		break;
	
	case RMASK:						/* Read the MASK register.		*/
   		ustemp = dp->btfp_mask;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case WMASK:						/* Write to the MASK register.	*/
		if( !( (int)arg & ( (int) usmask ) ) )
		{
			dp->btfp_mask = (ushort)arg; /* Directly from the user's argument.*/
									/* No interrupt enable requested.		*/
		}
		else if( !sp->signalon )	/* Signal sending is NOT enabled.		*/
		{
			dvrdisable(s, DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);	
									/* Block interrupts */

										
			if ( (usmask & MASK_DPA)			/* DPA already? */
				&& (dp->btfp_mask & (ushort) arg) ) /* and user wants DPA */
			{
				if ( !(sp->gps_synch) )				/* Are we in synch with GPS? */
				{
					dp->btfp_ack = ACK_CLEAR;			/*  Clear OUTFIFO			*/
													/* When we get the next */
													/* DPA, the interrupt   */
													/* routine will get us  */
													/* synch-ed up with GPS */
													/* and set the flag     */
				}
				dp->btfp_ack = ACK_DPA;			/* Clear DPA status 		*/
			} /* end of "DPA-already" section */

			sp->intwait = (ushort)arg;		/* Indicate we're waiting for an int.	*/
			dp->btfp_intstat = (ushort)arg;	/* Clear any existing interrupt */
									/* Save the interrupts we're waiting for */
			usmask = dp->btfp_intstat;				/* Clear ALL int. status */
			dp->btfp_intstat = usmask;				/* ... */
			dp->btfp_intstat = 0xFF;
			dp->btfp_mask = (ushort)arg; /* Enable bc63{57} interrupt(s).		*/

			dvrrestore(s,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);

			while ( sp->intwait )		/* btfp_int() will clear this on int. */
			{
										/* block, waiting for 637 interrupt. */
				swait(&DrvrData->card_data[minor(mmdev)].user_sem,0);
										/* Wait till our interrupt occurs. */
			}
		}
		else						/* Signal sending IS enabled.			*/
		{
			dvrdisable(s,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
									/* Block interrupts */
			if (usmask & MASK_DPA)			/* DPA already? */
			{
				if ( !(sp->gps_synch) )				/* Are we in synch with GPS? */
				{
					dp->btfp_ack = ACK_CLEAR;			/*  Clear OUTFIFO			*/
													/* When we get the next */
													/* DPA, the interrupt   */
													/* routine will get us  */
													/* synch-ed up with GPS */
													/* and set the flag     */
				}
				dp->btfp_ack = ACK_DPA;			/* Clear DPA status 		*/
			} /* end of "DPA-already" section */
			sp->intwait = (ushort)arg;		/* We will not be blocking*/
			dp->btfp_intstat = (ushort)arg;		/* Clear any existing interrupt */
			usmask = dp->btfp_intstat;				/* Clear ALL int. status */
			dp->btfp_intstat = usmask;				/* ... */
			dp->btfp_mask = (ushort)arg; /* Enable bc63{57} interrupt(s) etc.	*/
			dvrrestore(s,DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);
									/* Allow interrupts from our device again.*/
		}
		break;
	
	case RINTSTAT:						/* Read the INTSTAT register.		*/
   		ustemp = dp->btfp_intstat;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case WINTSTAT:						/* Write to the INTSTAT register.	*/
   		dp->btfp_intstat = (ushort)arg;/* Directly from the user's argument.	*/
		break;
	
	case RVECTOR:						/* Read the VECTOR register.		*/
   		ustemp = dp->btfp_vector;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;
	
	case RLEVEL:						/* Read the LEVEL register.		*/
   		ustemp = dp->btfp_level;		
		uitemp = 0x00ff & ustemp;	/* Preserve byte value only 			*/
		memcpy (arg, &uitemp, sizeof(int));	/* Move to caller's space */
		break;

	case READTIME:					/* Capture and read all time words.		*/
		ustemp		 		= dp->btfp_timereq;		/* Capture time.		*/
		tp->btfp_time[ 0 ] 	= dp->btfp_time[ 0 ];	/* Time word 0.			*/
		tp->btfp_time[ 1 ] 	= dp->btfp_time[ 1 ];	/* Time word 1.			*/
		tp->btfp_time[ 2 ] 	= dp->btfp_time[ 2 ];	/* Time word 2.			*/
		tp->btfp_time[ 3 ] 	= dp->btfp_time[ 3 ];	/* Time word 3.			*/
		tp->btfp_time[ 4 ] 	= dp->btfp_time[ 4 ];	/* Time word 4.			*/
		memcpy(arg,tp,sizeof(struct btfp_time));	
		break;	

	case READEVENT:					/* Read all event words.		*/
		ep->btfp_event[ 0 ] = dp->btfp_event[ 0 ];	/* Event word 0.		*/
		ep->btfp_event[ 1 ] = dp->btfp_event[ 1 ];	/* Event word 1.		*/
		ep->btfp_event[ 2 ] = dp->btfp_event[ 2 ];	/* Event word 2.		*/
		ep->btfp_event[ 3 ] = dp->btfp_event[ 3 ];	/* Event word 3.		*/
		ep->btfp_event[ 4 ] = dp->btfp_event[ 4 ];	/* Event word 4.		*/
		memcpy (arg, &btfpevent, sizeof (struct btfp_event));
		break;

	case WRITESTROBE:					/* Write to the Strobe registers.	*/
		memcpy (strp, arg, sizeof (struct btfp_strobe) );
		dp->btfp_cr0 = STRDIS;			/* First disable strobe output		*/
		dp->btfp_strobe1 = strp->btfp_strobe[0];	/* Strobe word 1		*/
		dp->btfp_strobe2 = strp->btfp_strobe[1];	/* Strobe word 2		*/
		dp->btfp_strobe3 = strp->btfp_strobe[2];	/* Strobe word 3		*/
		break;

	case READALLREGS:				/* Read all bc63{57} regs.				*/
		up->btfp_id		= dp->btfp_id;			/* ID register.		*/
		up->btfp_device = dp->btfp_device;		/* Device Type register.	*/
		up->btfp_status	= dp->btfp_status;		/* Status register.		*/

		up->btfp_timereq	= dp->btfp_timereq;		/* Capture the time.	*/
		up->btfp_time[ 0 ] 	= dp->btfp_time[ 0 ];	/* Time word 0.			*/
		up->btfp_time[ 1 ] 	= dp->btfp_time[ 1 ];	/* Time word 1.			*/
		up->btfp_time[ 2 ] 	= dp->btfp_time[ 2 ];	/* Time word 2.			*/
		up->btfp_time[ 3 ] 	= dp->btfp_time[ 3 ];	/* Time word 3.			*/
		up->btfp_time[ 4 ] 	= dp->btfp_time[ 4 ];	/* Time word 4.			*/

		up->btfp_event[ 0 ] = dp->btfp_event[ 0 ];	/* Event word 0.		*/
		up->btfp_event[ 1 ] = dp->btfp_event[ 1 ];	/* Event word 1.		*/
		up->btfp_event[ 2 ] = dp->btfp_event[ 2 ];	/* Event word 2.		*/
		up->btfp_event[ 3 ] = dp->btfp_event[ 3 ];	/* Event word 3.		*/
		up->btfp_event[ 4 ] = dp->btfp_event[ 4 ];	/* Event word 4.		*/

		up->btfp_unlock 	= dp->btfp_unlock;		/* UNLOCK register.	*/
	
		up->btfp_ack 		= dp->btfp_ack;			/* ACK register.	*/
		up->btfp_cr0 		= dp->btfp_cr0;			/* Control register 0.	*/
		up->btfp_outfifo	= dp->btfp_outfifo;		/* OUTFIFO register.	*/
		up->btfp_mask		= dp->btfp_mask;		/* MASK register.	*/
		up->btfp_intstat	= dp->btfp_intstat;		/* INTSTAT register.	*/
		up->btfp_vector		= dp->btfp_vector;		/* VECTOR register.	*/
		up->btfp_level		= dp->btfp_level;		/* LEVEL register.	*/
		memcpy (arg, up, sizeof (struct btfp_device));
	
		break;

	case SIGNALON:					/* Turn on signal sending on interrupts.*/
		sp->signalon = 1;			/* Indicate we should send signals on ints.*/
		break;
		
	case SIGNALOFF:					/* Turn off signal sending on interrupts.*/
		dvrdisable(s, DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);	
									/* Block interrupts */
		sp->signalon = 0;			/* Stop sending signals on interrupts.	*/
		sp->intwait = 0;			/* Indicate we are not awaiting any int.*/
		dp->btfp_mask = 0;			/* Turn off all 5 interrupts.	*/
		dvrrestore(s, DrvrData->card_data[minor(mmdev)].vector->interrupt_bit);	
									/* Restore interrupt */	
		break;

	case PACKETON:					/* Turn on packet mode for read/write.*/
		sp->packeton = 1;			
		break;

	case PACKETOFF:					/* Turn off packet mode for read/write.*/
		sp->packeton = 0;			
		break;
		
	case DEBUGON:
		sp->debugon = 1;			/* Enable driver debugging */
		break;
		
	case DEBUGOFF:
		sp->debugon = 0;			/* Disable driver debugging */
		break;

	default:							
		pseterr(EINVAL);			/* Return this is not a tty type device.*/
		return(SYSERR);
		break;
	}
	return( 0 );					/* Must return OK if no error found.	*/

} /* end of btfpioctl routine */

/*
  
btfpselect(DrvrData, fd, which ,se)
bcd *DrvrData;
struct file *fd;
int which;
struct sel *se;
{

	kprintf("\nNow in bcdrvselect\n");
	return(0);
}
 */ 
btfpuninstall(DrvrData)
bcd *DrvrData;
{
/* The Interrupt handler should be disabled here. */
  

/* deallocate memory used by static device information */
  
	sysfree (DrvrData, (long)sizeof (bcd));
	return (0);
}
  

/* btfp_int */

/* This interrupt routine does not do anything usefull.  It simply waits
   on a system semaphore, and writes a message to the console when an
   interrupt occurs.  The two things it does that are necessary are to
   clear the interrupt bit, and to turn interrupts back on for the device
   we are using. */

static void btfp_int(DrvrData,i)
register bcd *DrvrData;
register int i;
{
	register BTFPVOLATILE *dp = (BTFPVOLATILE *)
									DrvrData->card_data[i].pBC;
								/* Init pointer to base register */
	register struct btfp_status *sp = &btfp_status[i];
								/* Init pointer to status structure */
	unsigned short ustemp,ustemp2;	/* temps for register access */
	unsigned short usmask = MASK_INT0 | MASK_INT1 | MASK_INT2 |MASK_INT3 |MASK_INT4;
								/* Mask of module interrupt sources */

	while (1)
	{
		swait (DrvrData->card_data[i].int_info.sem, 0);
	 	usmask = MASK_INT0 | MASK_INT1 | MASK_INT2 |MASK_INT3 |MASK_INT4;
		/* Clear interrupt we received */
		dvrclearint (DrvrData->card_data[i].vector->interrupt_bit);
		/* Turn interrupts back on for this device */
		dvrenable (DrvrData->card_data[i].vector->interrupt_bit);

		ustemp = dp->btfp_intstat;		/* Read INTSTAT register */
		if( !(sp->intwait & ustemp) )	/* We weren't waiting for this int! */
		{
			dp->btfp_mask = 0;		/* Turn off all interrupt sources */
			dp->btfp_intstat = usmask;	/* Reset INTSTAT register */
			kprintf("btfpint(): Unexpected interrupt from btfp0\n");
			return;
		}
		else if( !sp->signalon)			/* We were waiting for an int.. */
		{								/* and were blocked doing so.   */
			sp->intwait = 0;			/* Indicate we received an int. */
			if (ustemp & INTSTAT_DPA )
			{
				/* For DPA interrupt, remember to clear the ACK_DPA bit */
				dp->btfp_ack = ACK_DPA;
			}

			dp->btfp_mask = 0;			/* Turn off all int sources  */
			dp->btfp_intstat = usmask;	/* Reset INTSTAT register    */
			ssignal(&DrvrData->card_data[i].user_sem); /* Send signal to blocked process */
		}
		else							/* Waiting for interrupt but in */
		{								/* SIGNALON mode...*/
			dp->btfp_intstat = usmask;	/* Reset INTSTAT register  */

			/* if DPA is here...*/
			if (ustemp & INTSTAT_DPA)
			{

			if (sp->debugon)
			{
				if (ustemp & INTSTAT_DPA)
					kprintf ("d");
			}

			/* ...AND the user wanted to look at it */
			if (sp->intwait & MASK_DPA)
			{
			if (sp->debugon)
			{
				if (ustemp & INTSTAT_DPA)
					kprintf("D");
			}

			dp->btfp_ack = ACK_DPA;		/* Clear DPA bit in ACK reg */

			/* NOTE: SIGNALON mode has already waited for the DPA signal.*/
			/* thus, its read just reads the available packet */

			/* Now fill the packet buffer from OUTFIFO */

			sp->rpacki = 0;

			/* We must read the first byte to activate the ACK_MORE flag */
			/* and to see if we are out of synch with the TANS packets   */
			/* If the first two bytes don't look like the start of a     */
			/* packet, just clear the OUTFIFO again                      */

			if (!sp->gps_synch)
			{

			sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo; /* Get 1rst byte*/
			if ( sp->rpack_buffer [sp->rpacki - 1] != (unsigned char)DLE )
			{
				/* Bogus packet...clear OUTFIFO */
				dp->btfp_ack = ACK_CLEAR;	/* Clear OUTFIFO */
			}
			else
			{
				/* First byte OK..now look at next byte; is it a packet ID? */
				sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;
				if ( ( sp->rpack_buffer [sp->rpacki - 1] == (unsigned char)DLE )
			  	 || (sp->rpack_buffer [sp->rpacki - 1] == (unsigned char)ETX ) )
				{
					/* Bogus packet...clear OUTFIFO */
					dp->btfp_ack = ACK_CLEAR;	/* Clear OUTFIFO */
				}
				/* else packet is OK....read it in */

			} /* end of first byte OK section */

			if (sp->debugon)
				kprintf ("S");
			sp->gps_synch = 1;					/* We're synched up now */

			} /* end of "need to synch up with the GPS" section */

			/* NOTE: ACK_MORE bit on means "more bytes in OUTFIFO" */

			ustemp = dp->btfp_ack;
			while ( (ustemp & ACK_MORE)			/* Until FIFO empty */
			  && (sp->rpacki < MAXPACKET) )		/* or running away */
			{
				sp->rpack_buffer [sp->rpacki++] = dp->btfp_outfifo;/*Get byte*/
				ustemp = dp->btfp_ack;			/* Any more? */

			} /* end of OUTFIFO read loop */

			if (sp->rpacki == 0)
			{
				if (sp->debugon)
					kprintf("0");
				/* Wait for next DPA interrupt */
				return;   
			}

			} /* end of "user wanted to see DPA" section */

			else
			{
				ustemp &= ~INTSTAT_DPA;		/* Any other interrupts happen? */
				if ( !(ustemp & sp->intwait) ) /*if not our interest, bail  */
				{
					return;
				}
				/* else, fall through to send the user a signal */
			}

			} /* end of DPA section */

			/* if no DPA, we must have matched on another INTSTAT bit */
			pkill(sp->procp,SIGUSR1);

		} /* end of SIGNALON section */

	} /* end of while loop, setup signal wait again */

} /* end of btfpint routine */

int send_packet(DrvrData, mmdev, packet, count)
bcd *DrvrData;
dev_t mmdev;
unsigned char *packet;
register int count;
{
	register BTFPVOLATILE *dp = (BTFPVOLATILE *) 
									DrvrData->card_data[ minor(mmdev) ].pBC;
								/* init pointer to base register */
	register struct btfp_status *sp = &btfp_status[ minor(mmdev) ];
								/* init pointer to status register */

	int i,loop_count;
	unsigned char *byte_ptr=packet;
	unsigned short ustemp;		/* for accessing ACK register */
	int	ps;						/* Interrupt level */

	dp->btfp_ack = ACK_INFIFO;	/* Clear ack bit */

	for(i=0;i<count;i++)
	/*while (*byte_ptr)*/		/* 6/15/95 gdowd */
	{
		dp->btfp_infifo = *byte_ptr;
		byte_ptr++;
	} /* end of while */


	dp->btfp_ack = ACK_INACT;		/* Take action bit */

									/* Wait for bc63{57} to take action */

									/* NOTE: no interrupt generated */
									/* so we have to spin */

	loop_count = 0;
	while (loop_count < MAX_SPIN)
	{
		ustemp = dp->btfp_ack;
		if ( ustemp & ACK_INFIFO ) /* Break if action taken */
		{
			break;
		}
		else
		{
			loop_count++;
		}
	}

	dp->btfp_ack = ACK_INFIFO;		/* Clear ack bit */

	if (loop_count >= MAX_SPIN)
	{
		kprintf("TIMEOUT (%d) waiting for INFIFO action\n",MAX_SPIN);
		return(SYSERR);				/* Return error */
	}
	else
	{
		if (sp->debugon)
			kprintf("%dLoop ", loop_count);
	}

	return(0);						/* Return no problem */

} /* end of send_packet routine */
